# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging

from driverng import MessageV2, Worklist
from driverng.fixed import Date, FixedLengthFormat, Multiple, String
from driverng.utils import num_years

logger = logging.getLogger("driverng.au5800")

SAMPLE_ID_LENGTH = 16


class Au5800Query(FixedLengthFormat):
    FMT_DESCRIPTION = [
        String("code", 2),
        String("unit_no", 2),
        String("rack", 4),
        String("cup", 2),
        String("sample_type", 1),
        String("sample_no", 4),
        String("sample_id", SAMPLE_ID_LENGTH, right_justified=True),
    ]


class Au5800Order(FixedLengthFormat):
    HEADER = [
        String("rack", 4),
        String("cup", 2),
        String("sample_type", 1),
        String("sample_no", 4),
        String("sample_id", SAMPLE_ID_LENGTH, right_justified=True),
        String("dummy", 4),
        String("block_identification_number", 1),
    ]
    PATIENT = [
        String("sex", 1),
        String("years_age", 3),
        String("months_age", 2),
        # Total of 6, we only use the first two ones.
        # The first for IPP and second for identity.
        String("patient_ipp", 20),
        String("patient_identity", 20, truncable=True),
    ]
    ORDER = [Multiple("tests", [String("code", 3), String("diluent_type", 1)])]

    FMT_DESCRIPTION = [String("code", 2)] + HEADER + PATIENT + ORDER


class Au5800RerunOrder(FixedLengthFormat):
    FMT_DESCRIPTION = [String("code", 2)] + Au5800Order.HEADER + Au5800Order.ORDER


class Au5800Result(FixedLengthFormat):
    TEST = [
        String("unit", 1),
        String("cuvette", 1),
        String("test_number", 3),
        String("reagent", 16),
        String("analysis_data", 9),
        String("data_flags", 8),
    ]
    FMT_DESCRIPTION = (
        [String("code", 2), String("unit_no", 2)]
        + Au5800Order.HEADER
        + Au5800Order.PATIENT
        + [Date("loading_time", 14)]
        + [Multiple("tests", TEST)]
    )


class Au5800RerunResult(FixedLengthFormat):
    FMT_DESCRIPTION = (
        [String("code", 2), String("unit_no", 2)]
        + Au5800Order.HEADER
        + [Date("loading_time", 14)]
        + [Multiple("tests", Au5800Result.TEST)]
    )


class Au5800QC(FixedLengthFormat):
    FMT_DESCRIPTION = (
        [String("code", 2), String("unit_no", 2)]
        + Au5800Order.HEADER
        + [Date("loading_time", 14)]
        + [Multiple("tests", Au5800Result.TEST)]
    )


class Driver(MessageV2):
    SPEC = {"code": "au5800", "label": "AU5800"}

    @staticmethod
    def get_klass(code):
        if code in ("Sh", "SH"):
            return Au5800RerunOrder
        elif code[0] == "S":
            return Au5800Order
        elif code[0] == "R":
            return Au5800Query
        elif code in ("DQ", "DR", "DA"):
            return Au5800QC
        elif code in ("Dh", "DH"):
            return Au5800RerunResult
        elif code[0] == "D":
            return Au5800Result
        else:
            raise NotImplementedError

    @staticmethod
    def get_reagents(data):
        reagents = []
        for i in range(0, 16, 4):
            reagent = data[i:][:4]
            if reagent:
                reagents.append(reagent)
        return reagents

    def create(self, code, **kwargs):
        kwargs.setdefault("code", code)
        return self.get_klass(code)(**kwargs)

    def parse(self, data):
        obj = self.get_klass(data[:2])()
        obj.parse(data)
        return obj

    def to_string(self, message):
        return message.to_string()

    def encode_worklist(self, worklist, context):
        samples = worklist.get_worklist(
            unique_together=["action"],
            action={"del": worklist.IGNORE, worklist.ELSE: None},
        )
        if len(samples) == 0:
            if not context["query"]:
                return
            else:
                query = context["query"]
                return self.create(
                    "S{}".format(query.code[1:]),
                    sample_type=" " if query.sample_type == "N" else query.sample_type,
                    sample_id=query.sample_id,
                    block_identification_number="E",
                    rack=query.rack,
                    cup=query.cup,
                    sample_no=query.sample_no,
                )
        elif len(samples) > 1:
            raise RuntimeError("Multi-sample not implemented")
        else:
            sample = samples[0][0]

        if sample.specimen == "Serum" or sample.specimen == "S":
            sample_type = " "
        else:
            sample_type = sample.specimen[0]

        query = context["query"]
        if query:
            code = "S{}".format(query.code[1:])
        else:
            code = "S"

        message = self.create(
            code,
            sample_type=sample_type,
            sample_id=sample.sample_id,
            block_identification_number="E",
        )
        if isinstance(message, Au5800Order):
            message.sex = sample.sex if sample.sex in ("M", "F") else ""
            message.years_age = num_years(sample.birthdate, sample.collection_date)
            message.patient_ipp = sample.patient_id
            message.patient_identity = "{} {}".format(
                sample.last_name, sample.first_name
            )
        if query:
            message.rack = query.rack
            message.cup = query.cup
            message.sample_no = query.sample_no

        for order in sample.orders:
            message.tests.add(code=order.code, diluent_type="0")
        return message

    def decode_message(self, message):
        if isinstance(message, Au5800Query):
            worklist = Worklist(context="query")
            worklist.get_or_create(type="patient", sample_id=message.sample_id.strip())
            return worklist

        worklist = Worklist(context="result")
        if isinstance(message, Au5800Result) or isinstance(message, Au5800RerunResult):
            rack = ""
            if message.rack and message.cup:
                rack = "{}_{}".format(message.rack, message.cup)
            elif message.rack and not message.cup:
                rack = message.rack
            sample = worklist.get_or_create(
                type="patient",
                sample_id=message.sample_id,
                specimen=message.sample_type or "S",
                rack=rack,
            )
        elif isinstance(message, Au5800QC):
            sample = worklist.get_or_create(
                type="control",
                control_lot=message.sample_id,
                sample_id=message.sample_id,
                specimen=message.sample_type or "S",
            )
        else:
            raise NotImplementedError

        for test in message.tests:
            flags = []
            if test.data_flags:
                for flag in test.data_flags:
                    if flag.strip():
                        flags.append(flag)
            result = sample.results.add(
                code=test.test_number,
                value=test.analysis_data,
                date=message.loading_time,
                flags=flags,
            )
            lots = self.get_reagents(test.reagent)
            for lot in lots:
                result.lots.add(type="reagent", lot=lot)

        return worklist
